theory ISA_Class
  imports Isabelle_DOF.DOF_Core

begin

ML\<open>
structure ISA_class = 
struct

fun ML_isa_check_generic check thy (term, pos) =
  let val name = (HOLogic.dest_string term
                  handle TERM(_,[t]) => error ("wrong term format: must be string constant: " 
                                               ^ Syntax.string_of_term_global thy t ))
      val _ = check thy (name,pos)
  in  SOME term end;

fun check_identity _ (term, _, _) _ = SOME term
fun check_instance thy (term, _, pos) s =
  let
    val bname = Long_Name.base_name s;
    val qual = Long_Name.qualifier s;
    val class_name =
        Long_Name.qualify qual (String.extract(bname , String.size(DOF_core.doc_class_prefix), NONE));
    fun check thy (name, _)  =
      let
        val object_cid = case DOF_core.get_object_global name thy of
                          NONE =>  DOF_core.err ("No class instance: " ^ name) pos
                          | SOME(object) => #cid object
        fun check' (class_name, object_cid) =
          if class_name = object_cid then
            DOF_core.get_value_global name thy
          else DOF_core.err (name ^ " is not an instance of " ^ class_name) pos
      in check' (class_name, object_cid) end;
  in ML_isa_check_generic check thy (term, pos) end

fun elaborate_instance thy _ _ term_option pos =
  case term_option of
      NONE => error ("Malformed term annotation")
    | SOME term => let val instance_name = HOLogic.dest_string term
                   in case DOF_core.get_value_global instance_name thy of
                          NONE => error ("No class instance: " ^ instance_name)
                        | SOME(value) =>
                            DOF_core.transduce_term_global {mk_elaboration=true} (value, pos) thy
                   end

(*
  The function declare_ISA_class_accessor_and_check_instance uses a prefix
  because the class name is already bound to "doc_class Regular_Exp.rexp" constant
  by add_doc_class_cmd function
*)
fun declare_ISA_class_accessor_and_check_instance doc_class_name =
  let
    val bind = Binding.prefix_name DOF_core.doc_class_prefix doc_class_name
    val typestring = "string => " ^ (Binding.name_of doc_class_name)
    (* Unfortunately due to different lexical conventions for constant symbols and mixfix symbols
       we can not use "_" for classes names in term antiquotation.
       We chose to convert "_" to "-".*)
    val conv_class_name = String.translate (fn #"_" => "-"
                                            | x => String.implode [x] )
                                                (Binding.name_of doc_class_name)
    val mixfix_string = "@{" ^ conv_class_name ^ " _}"
  in
    Sign.add_consts_cmd [(bind, typestring, Mixfix.mixfix(mixfix_string))]
    #> (fn thy => let
                    val long_name = DOF_core.read_cid_global thy (Binding.name_of doc_class_name)
                    val qual = Long_Name.qualifier long_name
                    val class_name = Long_Name.qualify qual
                            (DOF_core.get_doc_class_name_without_ISA_prefix (Binding.name_of bind))
                  in
                    DOF_core.update_isa_global
                             (class_name, {check=check_instance, elaborate=elaborate_instance}) thy
                  end)
  end

fun elaborate_instances_list thy isa_name _ _ _ =
  let
    val base_name = Long_Name.base_name isa_name
    fun get_isa_name_without_intances_suffix s =
      String.extract (s, 0, SOME (String.size(s) - String.size(instances_of_suffixN)))
    val base_name_without_suffix = get_isa_name_without_intances_suffix base_name
    val base_name' = DOF_core.get_class_name_without_prefix (base_name_without_suffix)
    val class_typ = Proof_Context.read_typ (Proof_Context.init_global thy)
                                                (base_name')
    val tab = #tab(#docobj_tab(DOF_core.get_data_global thy))
    val table_list = Symtab.dest tab
    fun get_instances_name_list _ [] = []
      | get_instances_name_list class_name (x::xs) =
          let 
            val (_, docobj_option) = x
          in 
            case docobj_option of
                NONE => get_instances_name_list class_name xs
              | SOME {cid=cid, value=value, ...} =>
                  if cid = class_name
                  then value::get_instances_name_list class_name xs
                  else get_instances_name_list class_name xs
          end
    val long_class_name = DOF_core.read_cid_global thy base_name'
    val values_list = get_instances_name_list long_class_name table_list
  in HOLogic.mk_list class_typ values_list end

fun declare_class_instances_annotation thy doc_class_name =
  let
    val bind = Binding.prefix_name DOF_core.doc_class_prefix doc_class_name
    val bind' = Binding.suffix_name instances_of_suffixN bind
    val class_list_typ = Proof_Context.read_typ (Proof_Context.init_global thy)
                                                ((Binding.name_of doc_class_name) ^ " List.list")
    (* Unfortunately due to different lexical conventions for constant symbols and mixfix symbols
       we can not use "_" for classes names in term antiquotation.
       We chose to convert "_" to "-".*)
    val conv_class_name' = String.translate (fn #"_" => "-" | x=> String.implode [x])
                                        ((Binding.name_of doc_class_name) ^ instances_of_suffixN)
    val mixfix_string = "@{" ^ conv_class_name' ^ "}"
  in
    Sign.add_consts [(bind', class_list_typ, Mixfix.mixfix(mixfix_string))]
    #> (fn thy => let
                    val long_name = DOF_core.read_cid_global thy (Binding.name_of doc_class_name)
                    val qual = Long_Name.qualifier long_name
                    val transformer_name = Long_Name.qualify qual
                                (DOF_core.get_doc_class_name_without_ISA_prefix (Binding.name_of bind'))
                  in
                    DOF_core.update_isa_global (transformer_name,
                    {check=check_identity, elaborate= elaborate_instances_list}) thy end)
  end

end (* ISA_class *)
\<close>

end